{-# LANGUAGE LambdaCase #-}

-- | Recursive decent (LL) parsing producing abstract syntax.

-- Left-factored grammar:

-- S  ::= Integer S';
-- S' ::= ;
-- S' ::= "+" S;
-- S' ::= "*" P S';
-- P  ::= Integer P' ;
-- P' ::= ;
-- P' ::= "*" P ;

data Token
  = TInt Integer
  | TPlus
  | TTimes

-- We parse a sum of products (this is our abstract syntax).

type Sum  = [Prod]
type Prod = [Integer]

-- Parsers for sum.

parseS :: [Token] -> Maybe Sum
parseS = \case
  TInt i : rest -> parseS' [i] rest
  _ -> Nothing

parseS' :: Prod -> [Token] -> Maybe Sum
parseS' p = \case
  []            -> return [p]

  TPlus  : rest -> do
    s <- parseS rest
    return (p : s)

  TTimes : rest -> do
    (p', rest') <- parseP rest
    parseS' (p ++ p') rest'

  _ -> Nothing

-- Parsers for product.

parseP :: [Token] -> Maybe (Prod, [Token])
parseP = \case

  TInt i : rest -> do
    (p, rest') <- parseP' rest
    return (i:p, rest')

  _ -> Nothing

parseP' :: [Token] -> Maybe (Prod, [Token])
parseP' = \case
  TTimes : rest -> parseP rest
  inp -> return ([], inp)

-- General type of a
--
--   * deterministic parser: at most one result
--
--       [t] -> Either ParseError (a, [t])
--
--   * non-deterministic parser: 0 or more results
--
--       [t] -> [ (a, [t]) ]
--
-- Further reading: Parser Combinators

-- Testing

yes = parseS [TInt 1, TPlus, TInt 2, TTimes, TInt 3, TPlus, TInt 4]   -- 1 + 2 * 3 + 4
no  = parseS [TInt 1, TPlus, TPlus, TInt 2]                           -- 1 + + 2
